#ifndef DDEGIO_H
#define DDEGIO_H

#ifndef NO_VBX_SUPPORT
#define NO_VBX_SUPPORT
#endif
                                     
#include <ddeml.h> 

/*---------------------------------------------------------------------------                                     

DDE Classes by G. Bavestrelli              Copyright Techint S.p.A. 1995-1997

This module contains various classes that encapsulate DDEML library functions.
It is possible to exploit most of DDE's functionality easily with these classes.

The main two classes are DDEClient and DDEServer
------------------------------------------------

DDEClient encapsulates all the functionality of a DDE Client conversation.
You will need a DDEClient object for each conversation you want to be active concurrently.
With one DDEClient object you can have many conversations, but one at a time.
One conversation does not mean one transaction. You can have any number of
transactions (synchronous and/or asynchronous) with just one DDEClient object.

DDEServer encapsulates all the functionality of a DDE Server service.
You can have only one object inherited from DDEServer in your application.
Such an object can serve a single service at a time.
This is a limit suggested by Microsoft as standard DDE practice.
The service name is usually the name of your application.


You can use these classes in one of two ways:
1) Inherit from them and redefine the necessary virtual member functions.
   You can use multiple inheritance and, for example, inherit your CDocument,
   CMainFrame or CWinApp object from DDEClient and/or DDEServer also. 
2) You can define a class inherited from DDEClient or DDEObject and put an
   object of such a class anywhere. This object will handle DEE conversations.
   For client conversation you can use a DDEClient object direcltly.

You do not need to inherit from DDEClient if you use only synchronous 
transactions and you do not set up any advise link. In such a case, you can
use a DDEClient object directly. But if you want to use asynchronous transactions
and/or set up warm/hot advise links, you must inherit from DDEClient and 
redefine the necessary virtual member functions.

To use DDEServer you always need to inherit from DDEServer and redefine some
virtual member functions.


There are some other little classes you need to work with:
1) Class DDEString
2) Class DDEDataIn
3) Class DDEDataOut
4) Class DDEDataApp

DDEString
---------
DDEString is an abstraction of a DDE string handle (HSZ) and corresponding 
string (char *)
You can use a DDEString object anywhere a LPCSTR is expected and anywhere a
HSZ is expected, as it has the necessary conversion functions.
You can also construct a DDEString object from any string or string handle.
Some DDEString objects are also passed to the virtual functions you might redefine.
You never have to free a string handle if you create all your DDEString objects 
on the stack or as static or global variables (recommended).
Otherwise, if you create them on the heap, you'll naturally have to delete them.
DDEString class has many operators to test its value: look at the class definition
which follows in this H file.
Note that you do not have to create any DDEString objects, as you can always pass
strings instead. 
But if you keep DDEString objects you use often, you'll increase performance.
You can pass a LPSTR or an HSZ to any function that takes a const DDEString &.

DDEDataIn
---------
When you receive some data from DDE, you will receive it inside a DDEDataIn object.
For example, if you want to do a DDE Request, you'll have to create a DDEDataIn
object (probably on the stack) and pass it to the Request member function.
If such a function will succeed, inside the DDEDataIn object you'll have the data.
You can access this data with DDEDataIn's member functions. (see class definition).

DDEDataOut
----------
When you send some data through DDE, you will need to fill a DDEDataOut object
and pass it to some member function of DDEClient or DDEServer.
To fill such an object with data you will use its member functions.
See the class definition below.

DDEDataApp
----------
Abstraction of HDATA_APPOWNED data, which is kept by application and released
only when object goes out of scope. You can pass such data many times through DDE.

Note on DDEDataIn and DDEDataOut
--------------------------------
Always create these objects on the stack, you will never have to free any of them!!!


See the comments below for the various functions.

---------------------------------------------------------------------------*/


// Macros to handle sync/async transactions
#define SYNC_TRAN 0xFFFFFFFF        // Identifies synchronous transactions
#define ASYNC(a)  ((a)!=SYNC_TRAN)  // Is it an Id specifying an asynch trans?

// Parameter to pass as default to indicate zero terminated string as data
#define ZERO_TERMINATED 0xFFFFFFFF  // Length = zero terminated string


class DDEString;
class DDEData;
class DDEDataIn;
class DDEDataOut;
class DDEDataApp;
class DDEClient;
class DDEServer;


//---------------------------------------------------------------------------                                     
//---------------------------------------------------------------------------
// DDE Initialization functions
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
                                  
// Initialize DDE, must call it before any DDE function is called
// If you don't call it, it will get called when you create the first DDEClient or DDEServer object
BOOL DDEInitialize(DWORD flags=APPCLASS_STANDARD); 

// Uninitializes DDE, must call it after all DDE functions have finished
// If you don't call it, it will get called after tha last DDEClient or DDEServer object is destroyed
BOOL DDEUninitialize();



//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
// DDEString class, abstraction over DDE strings and string handles
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
                                                                             
// Note: A DDEString can be used everywhere a HSZ or a LPCSTR are expected!
//       See the conversion operators and the overloaded constructors!!!!!!                                                                             


// Class to store string and handle
class DDEString
{
  CString s;  // String content
  HSZ     h;  // String handle

 public:

  // Construct a DDEString in various ways 
  DDEString()                     :h(0) {           }
  DDEString(HSZ hsz)              :h(0) { Set(hsz); }
  DDEString(LPCSTR str)           :h(0) { Set(str); } 
  DDEString(const CString & cst)  :h(0) { Set(cst); }
  DDEString(const DDEString & dde):h(0) { Set(dde); }
  virtual ~DDEString()                  { Reset() ; }
  
  // Setting and resetting strings
  void Set(HSZ handle);
  void Set(LPCSTR str);
  void Set(const CString & str);
  void Set(const DDEString & str);
  
  void Reset(); // Free and clear
  BOOL IsSet() const { return h || !s.IsEmpty(); }

  // Get information about DDEString
  LPCSTR GetString() const; 
  HSZ    GetHandle() const; 

  // Operators to enable passing DDEStrings in place of HSZ or LPCSTR
  operator LPCSTR () const { return GetString(); } 
  operator HSZ ()    const { return GetHandle(); }
  
  // Test for equality of two strings (not case sensitive)
  BOOL operator == (HSZ hsz)    const;
  BOOL operator == (LPCSTR str) const;            
  BOOL operator == (const CString & cst)   const;                                    
  BOOL operator == (const DDEString & dde) const;                                 

  BOOL operator != (HSZ hsz)    const            { return ! operator ==(hsz); }
  BOOL operator != (LPCSTR str) const            { return ! operator ==(str); }            
  BOOL operator != (const CString & cst)   const { return ! operator ==(cst); }                                    
  BOOL operator != (const DDEString & dde) const { return ! operator ==(dde); }                                 
  
  // Assignment operators 
  DDEString & operator = (HSZ hsz)               { Set(hsz); return *this; }
  DDEString & operator = (LPCSTR str)            { Set(str); return *this; }
  DDEString & operator = (const CString & cst)   { Set(cst); return *this; }  
  DDEString & operator = (const DDEString & dde) { Set(dde); return *this; }
  
  // Friend inline functions
  inline friend operator == (HSZ hsz            , const DDEString & dde) { return dde==hsz; }
  inline friend operator == (LPCSTR str         , const DDEString & dde) { return dde==str; }
  inline friend operator == (const CString & cst, const DDEString & dde) { return dde==cst; }
                            
  // Friend inline functions
  inline friend operator != (HSZ hsz            , const DDEString & dde) { return dde!=hsz; }
  inline friend operator != (LPCSTR str         , const DDEString & dde) { return dde!=str; }
  inline friend operator != (const CString & cst, const DDEString & dde) { return dde!=cst; }
                                              
  // Check if something is contained                                            
  BOOL IsValid() const { return !s.IsEmpty() || h;}
  BOOL IsEmpty() const { return s.IsEmpty && !h  ;}   

 private:
 
  BOOL HasString() const { return !s.IsEmpty(); }
  BOOL HasHandle() const { return h!=NULL     ; }  
  
};


//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// DDEData class, abstraction over DDE data handles and DDE data
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------



// Base Class to store DDEData
class DDEData
{
  protected:

   HDDEDATA hData;     // Data Handle  
   UINT     format;    // Data format
   HSZ      item;      // Item string handle

  protected:

   DDEData();	       // Create empty object
   virtual ~DDEData();

  public:

   virtual HDDEDATA GetHandle()  const { return hData; } // Get DDE data handle
   virtual UINT     GetFormat()  const { return format;} // Get clipboard format 
   virtual HSZ      GetItem()    const { return item;  } // Get item of data
   
  protected:

   virtual void Reset(); // Reset state of object to empty, free memory (usually called automatically)
   virtual void Free(){}

  private:

   operator = (const DDEData & d); // Disable equal operator
   DDEData(const DDEData &); // Disable copy constructor
};


//---------------------------------------------------------------------------
// You use a DDEDataIn object each time you receive data from DDE
//---------------------------------------------------------------------------

// Class to store DDEData that is passed to application from DDE
class DDEDataIn : public DDEData
{  
   LPVOID   pData;  // Pointer to data
   DWORD    nData;  // Length of data
   
   BOOL     toUnaccess;// True if we must call DdeUnaccessData
   BOOL     toFree;    // True if we must call DdeFreDataHandle
   
  public:

   DDEDataIn();
  ~DDEDataIn();

   LPVOID GetData() const   { return pData; } // Access the data
   DWORD  GetSize() const   { return nData; } // Get data size


   // Helper if you know the data is a string
   LPCSTR GetString() const;


   void Reset(); // Reset state of object to empty, free memory (usually called automatically)     
         
  protected: // The following are used only by the DDE classes

   // Constructor used when receiving data in callback
   DDEDataIn(HDDEDATA handle,UINT format,HSZ item);
   // Function used when DdeClientTransaction returns data
   BOOL SetForClientRequestTransaction(HDDEDATA handle,UINT format,HSZ item);
   // Free data handle, keep format and item
   void Free(); 
   
   // Friends
   friend class DDEClient;
   friend class DDEServer;
};


//---------------------------------------------------------------------------
// You use a DDEDataOut object each time you send data to DDE
//---------------------------------------------------------------------------

// Class to store DDEData to be passed to DDE from application
class DDEDataOut : public DDEData
{             
   BOOL PassedToDDEML;         // True if data handle was passed to DDEML
   DDEDataApp * pDataAppOwned; // Points to DDEDataApp object that contains handle

  public: 

   // Set data from an application owned data object                                   
   BOOL SetDataApp(DDEDataApp & Appdata);
   
   // Call this function to set the new data to send to DDE
   // If you pass a string, pass its length + 1 for zero terminator
   BOOL SetData(LPCVOID pdata, DWORD lendata);     
   // Adds data to the object starting at specified offset
   BOOL AddData(LPCVOID pdata, DWORD lendata, DWORD offset);


   // Use this method to set data if data is a string
   BOOL SetString(LPCSTR str);

   
   // Reset state of object to empty, free memory                                    
   void Reset(); 

  protected:

   void Free();
   void SetPassedToDDEML(); 
   HDDEDATA GetHandle() const; 
          
   // Protected constructor: only DDE Classes can create these       
   DDEDataOut(UINT format, HSZ item);
  ~DDEDataOut();
   
   friend class DDEServer;
};

//---------------------------------------------------------------------------
// You use a DDEDataApp object to keep a data handle in your app: HDATA_APPOWNED
//---------------------------------------------------------------------------


// Class to store DDEData to be passed to DDE from application
class DDEDataApp : public DDEData
{                                         
   BOOL PassedToDDEML; // TRUE if handle was passed to DDEML
   int  CountReferences;

  public: 

   DDEDataApp();
  ~DDEDataApp();

   // Call this function to set the data to send to DDE
   // Remember to pass length of string + 1 for terminator if you pass a string
   BOOL SetData(LPCVOID pdata, DWORD lendata, const DDEString & item, UINT format);     
   // Adds data to the object starting at specified offset
   BOOL AddData(LPCVOID pdata, DWORD lendata, DWORD offset);

   void Reset();                    
   
  private:

   void Free();  
   void SetPassedToDDEML(); 
   void IncrReferenceCount(int incr);
   
   friend class DDEDataOut;
};


//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
// Functions that don't belong to DDEClient or DDEServer but affect both
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

  // Suspends all transactions in all conversations
  BOOL DDESuspendAllConvTransactions();
  // Resumes all transactions or just one from the lot
  BOOL DDEResumeAllConvTransactions(BOOL JustOne);


//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
// DDEClient class, abstraction over DDE client conversations
// You can use these objects directly of inherit from them
// You will need to inherit if you use asynchronous transactions or advise links
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

// This is the abstraction over a client conversation
class DDEClient 
{
   HCONV hConversation;       // Handle of DDE conversation
   UINT  TimeOut;             // Timeout for transactions

   DDEString Service;	      // Service name
   DDEString Topic; 	      // Topic name
   
   static CMapPtrToPtr ClientMap; // Maps conv handles to DDEClient objects
   static CPtrList ClientList;    // List of all DDEClient objects

   BOOL MustReportErrors;     // TRUE if errors must be passed to reporting function
   BOOL MustTryDdeReconnect;  // TRUE if we must try reconnecting upon disconnection

   int CountAsynchTransPending; // Count pending asynch transactions
   int CountSynchTransPending;  // Count synchronous transactions pending (should be 0 or 1)

   BOOL TryingToConnect;        // TRUE if we are trying to connect (to overcome NETDDE problem)

 public:

   // Constructor and destructor
   DDEClient();
   virtual ~DDEClient();

   // Client conversation connection, use to start/stop/test connections
   BOOL Connect(const DDEString & service, const DDEString & topic); // Connects to service/topic unless already connected there
   BOOL Disconnect();                          // Disconnects from service
   
   BOOL IsConnected() const;                   // Returns true if it is connected
   BOOL IsConnected(const DDEString & service, const DDEString & topic) const;

   BOOL IsTryingToConnect() const { return TryingToConnect; }

   // Functions that start client transactions follow:
   // If you pass AsyncUserID!=SYNC_TRAN the transaction will be asynchronous.
   // In that case the AsyncUserID that you pass will be passed to TransactionCompleted
   //   when the asynchronous transaction is completed.
   // When an asynchronous transaction is completed, the virtual function
   //   TransactionCompleted will be called with the result and the data of the 
   //   transaction.    
   // For the meaning of each transaction see DDE documentation.

   // Start a request transaction. Pass a DDEDataIn object where the data will be put.
   // For asynchronous transactions, you can pass NULL as the DDEDataIn pointer. 
   BOOL  Request(const DDEString & Item,DDEDataIn * Data,UINT format=CF_TEXT,DWORD AsyncUserID=SYNC_TRAN, LPDWORD pdwResult=NULL);

   // Send unsolicited data to the server. The data is pointed to by ptrData.
   BOOL  Poke(const DDEString & Item, LPCVOID ptrData, DWORD length=ZERO_TERMINATED, UINT format=CF_TEXT,DWORD AsyncUserID=SYNC_TRAN, LPDWORD pdwResult=NULL);

   // Send a DDE execute string to the server
   BOOL  Execute(LPCSTR sCommand,DWORD AsyncUserID=SYNC_TRAN, LPDWORD pdwResult=NULL);

   // Start an advise link.
   BOOL  StartAdvise(const DDEString & Item,BOOL Warm=FALSE,BOOL ackRequired=FALSE,UINT format=CF_TEXT,DWORD AsyncUserID=SYNC_TRAN, LPDWORD pdwResult=NULL);

   // Stop an advise link.
   BOOL  StopAdvise(const DDEString & Item, UINT format=CF_TEXT,DWORD AsyncUserID=SYNC_TRAN, LPDWORD pdwResult=NULL);

      

   // Abandons specified transaction or all transactions of current conversation
   BOOL  AbandonAsyncTransaction(DWORD TransactionId=NULL); // NULL=All                    
   // Abandons all transactions of all conversations
   static BOOL AbandonAllAsyncTransactions();      
        
   // Suspends transactions of current conversation                                             
   BOOL SuspendConvTransactions();
   // Resumes all or just one transaction of current conversation
   BOOL ResumeConvTransactions(BOOL JustOne);
                                                
                                                
   // Set and get client transaction timeouts
   void  SetTimeout(UINT t)          { TimeOut=t;      }
   UINT  GetTimeout() const          { return TimeOut; }

   // Get the service and topic of current conversation
   const DDEString & GetService() const { return Service; }
   const DDEString & GetTopic()   const { return Topic;   }
                                                        
                                                        
   // Virtual functions to redefine to handle DDE events 
   // Redefine them to do what you want with the events:

   // Called when server sends us advise data (data=NULL for warm link)
   virtual BOOL ReceiveAdvise(const DDEString & Item, const DDEDataIn * const data, UINT format)
       {return FALSE;} // Return TRUE if you accept the data
   
   // Called when an async transaction has completed (data=NULL if no data is passed)
   virtual void TransactionCompleted(const DDEString & Item,const DDEDataIn * const data,UINT format,DWORD UserID,BOOL Ok)
       {}
   
   // Called when we got disconnected by the server
   virtual void Disconnected()
       {}
   
   // Called when services are registered/unregistered in system. Register=FALSE for Unregister
   virtual void Register(const DDEString & BaseServiceName, const DDEString & InstanceServiceName, BOOL Register)
       {}
   
   // Called when a DDE memory error occurs, release your data and close conversations if possible
   // Can get passed the handle of the conversation causing the error or NULL
   virtual void OnError(HCONV hConvError)
       {}

   // Return the handles relative to conversation and DDE
   HCONV GetConversationHandle() const { return hConversation; } // Handle to conversation

   static DWORD GetDDEHandle();        // Handle to DDE returned by DdeInitialize 

   // Get counter of uncompleted asynchronous transactions
   int AsynchBusy() const { return CountAsynchTransPending>=0?CountAsynchTransPending:0; }
   // Get counter of synch transactions uncompleted 
   int SynchBusy()  const { return CountSynchTransPending>=0 ?CountSynchTransPending :0; }

   // Return TRUE if any type of transaction is waiting to complete
   BOOL AnyTransBusy() const { return CountSynchTransPending>0 || CountAsynchTransPending>0; }

   // Count the number of pending transactions: synch, asynch, or both
   // This routine counts all pending transactions for all clients in app
   static int CountPendingTransactions(BOOL Synch=TRUE, BOOL Asynch=TRUE);


   // Error handling function: override for your own behaviour
   // By default it calls void ReportDDEErrorToApp(LPCSTR message);                              
   virtual void ReportError(UINT LastError, LPCSTR message);
   BOOL ReportDDEError(LPCSTR message);
   void EnableDDEErrorReport(BOOL On) // Should report errors to app?
      { MustReportErrors=On;} 

   // If you set this, client will try reconnecting upon disconnection 
   // If set to enabled, TryReconnecting will be called upon disconnection
   void SetAutomaticReconnection(BOOL On)
      { MustTryDdeReconnect=On; }	    

   // Called when we got disconnected by server to try to reconnect
   // You will usually not redefine this function, 
   // but you can if you want to be notified of reconnection 
   // This function actually does the reconnecting, 
   // so remember to call it in your override
   virtual BOOL TryReconnecting();


   // This is just a shortcut function to handle simple text synchronous requests
   // It is not as general as the Request function defined above, it's just here for simplicity
   // It only handles the case the request is for text and it is synchronous
   // Note that it fails if the Request succeeds and the received text is empty
   // You cannot use it if an empty text is valid for you, use Request instead in that case
   // It works only if the data received is zero-terminated
   // This function is just a shortcut for a common use of the Request function
   BOOL SimpleTextRequest(const DDEString & Item,CString & String);

  private: // Private functions
   
   void HaveDisconnected(); // handle disconnection

   
   BOOL SetUserHandle(DWORD idTrans, DWORD appValue);
   BOOL GetUserHandle(DWORD idTrans, DWORD * appValue);

   friend HDDEDATA CALLBACK DDECallback(UINT,UINT,HCONV,HSZ,HSZ,HDDEDATA,DWORD,DWORD);   
   HDDEDATA DDEClientCallback(UINT,UINT,HCONV,HSZ,HSZ,HDDEDATA,DWORD,DWORD);

   operator = (const DDEClient &); // Disable
   DDEClient(const DDEClient &);   // Disable
};



//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
// DDEServer class, abstraction over DDE server object
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------


// DDE Server class, abstraction on DDE service                      
class DDEServer
{
   DDEString ServiceName; // The only service name
   DDEString SystemTopic; // The system topic

   BOOL bAcceptSelfConnections;      // True if server accepts self connections
   BOOL isOpen;                      // True if connection is open

   static DDEServer * theDDEServer; // Pointer to the DDE Server object: only one server is allowed
   static CPtrList ServerConversationList; // List of conversations of server

   BOOL MustReportErrors;     // TRUE if errors must be passed to reporting function

  public: 
   
   DDEServer(); // Create object
   virtual ~DDEServer();

   
   // Opens DDE service. You always have to call OpenServer first to start the server.
   BOOL OpenServer(const DDEString & ServiceName);

   // If you don't call CloseServer, it will get called by destructor.
   // Call it if you want to start object again with another service.
   BOOL CloseServer();

   // Return TRUE if the server is open
   BOOL IsOpen() const { return isOpen; }


   // Conversation virtual functions called by system to start/stop conversations
   // You can redefine them to override default behaviour.
   
   // You MUST redefine this member to accept conversations.
   // If you return TRUE, the conversation is accepted.
   virtual BOOL AcceptConversation(const DDEString & Topic){return FALSE;}
   
   // This member notifies you that a conversation has started and passes you the handle.
   // You usually don't need to redefine this member function and dont care about the handle.
   virtual void BeginConversation(const DDEString & Topic,HCONV hConversation){}

   // This gets called when the conversation is ended
   // You usually don't need to redefine this member function.
   virtual void EndConversation(const DDEString & Topic,HCONV hConversation){}



   // If you call this you stop the conversation.
   // You usually don't need to call this.
   BOOL TerminateConversation(HCONV hConv); // Server terminates it
   void TerminateAllConversations(); // Terminate them all but leave server open

   // Return the number of active conversations
   int GetNumActiveConversations() const { return ServerConversationList.GetCount(); }
 


   // Virtual functions called by system to complete transactions:
   // You can usually ignore the HCONV papameter.

   // This is called when a client requests some data from the server.
   // Set the data inside the passed DDEData object (the object must not be freed)
   virtual BOOL HandleRequest(const DDEString & topic, const DDEString & item, DDEDataOut & Data, HCONV hConv)
       { return FALSE; } // Return TRUE if you handled message

   // This is called when a client pokes some data to this server.
   // The data is inside the DDEDataIn object. return TRUE if you accept it.
   virtual BOOL HandlePoke(const DDEString & topic, const DDEString & item, const DDEDataIn & Data, HCONV hConv)
       { return FALSE; }// Return TRUE if you handled message


   // Called when server receives an execute transaction from a client.
   // The complete execute command is in the command parameter.
   // This method will call HandleCommand for each command in execute transaction.
   // You can either redefine this one and handle the complete execute command string,
   // or leave this as it is and override HandleCommand to process each single parsed command
   // Here in commands multiple DDE commands can arrive, they get split up and sent to HandleCommand individually.
   virtual BOOL HandleExecute(const DDEString & Topic, LPSTR commands, HCONV hConv)
		{ return FALSE; }
   
   // Virtual functions that handle warm/hot links.
   // They get called to implement advise links.

   // Return TRUE to handle advise loop: this just enables it, does not return data
   virtual BOOL AcceptAdvise(const DDEString & topic, const DDEString & item, UINT format, HCONV hConv)
       { return FALSE;}

   // This returns the data requested in the advise loop. 
   // CountRemaining contains num remaining transactions or CADV_LATEACK
   // CountRemaining can be equal to CADV_LATEACK if Advise had AckRequired==TRUE and client was late in aknowledging
   // Search DDE documentation for CADV_LATEACK
   virtual BOOL PrepareAdviseData(const DDEString & topic, const DDEString & item, DDEDataOut & Data, UINT CountRemaining, HCONV hConv)
       { return FALSE;}

   // Call when an advise request is terminated: don't need to do anything
   virtual void AdviseStop(const DDEString & topic, const DDEString & item, UINT format, HCONV hConv)
       {}


   // Called when registering/unregistering service names. Register=FALSE for Unregister
   virtual void Register(const DDEString & BaseServName, const DDEString & InstServName, BOOL Register)
       {}
   
   // Called when a DDE memory error occurs, release your data and close conversations if possible
   // Can get passed the handle of the conversation causing the error or NULL
   virtual void OnError(HCONV hConvError)
       {}

   // Tell DDE that a topic/item pair has changed.
   // You must call this function when an item changes.
   // The system will then call the appropriate virtual functions.
   BOOL PostAdvise(const DDEString & topic, const DDEString & item);
   
   // Tell DDE that all topic/item pairs have been changed                                                                 
   // You must call this function when all your data has changed. 
   // The system will then call the appropriate virtual functions.
   BOOL PostAdvise();                                                                    
                              
   
   // You can enable or disable the acceptance of self connections
   // They are disabled by default      
   void AcceptSelfConnections(BOOL accept) { bAcceptSelfConnections=accept; }                           
   BOOL AcceptSelfConnections() const      { return bAcceptSelfConnections; }                           

   static DWORD GetDDEHandle(); // Retrun handle to DDEML

   //Error handling function: override for your own behaviour
   virtual void ReportError(UINT LastError, LPCSTR message);
   BOOL ReportDDEError(LPCSTR message);
   void EnableDDEErrorReport(BOOL On) // Should report errors to app?
      { MustReportErrors=On;} 

                              
                              
   // System topic functions ----------------------------------
   
   // Handles system topic requests.
   // By default it calls the AddSystem... functions so you usually dont need to redefine it
   virtual BOOL HandleSystemTopicRequest(const DDEString & item,CString & SysTopicResponse);                           

   // These functions are called by HandleSystemTopicRequest to return system topic items.
   // You get passed a string to which you must ADD your system topic implementation.
   // You must separate the items you return with a tab (\t)
   // The string will not be passed empty, so you must add a \t before each item you add.
   // You can clear up the string passed to get total control of system topic.
   // Standard system topic is handled by default.
   virtual void SystemAddSysItems(CString & SysItems){}
   virtual void SystemAddFormats(CString & Formats){}
   virtual void SystemAddTopics(CString & Topics){}  

 private: // Private members


   friend HDDEDATA CALLBACK DDECallback(UINT,UINT,HCONV,HSZ,HSZ,HDDEDATA,DWORD,DWORD);   
   HDDEDATA DDEServerCallback(UINT,UINT,HCONV,HSZ,HSZ,HDDEDATA,DWORD,DWORD);

   operator = (const DDEServer &); // Disable
   DDEClient(const DDEServer &);   // Disable
};                      




//-----------------------------------------------------------------------------


// Start network DDE if you need it
BOOL StartNetworkDDE();


#endif


